"Bootstrap for closures.  This directory should contain

	install.st - this file, which files in the change sets comprising the bootstrap in the right order
	cleanup.st - script to clean-up unbound methods and BlockContexts to finish the bootstrap
	changesets - a directory containing the various changesets comprising the bootstrap

File in install.st (this file) to load the bootstrap and recompile the system.  If you are prompted that a class variable in CompiledMethod is still used, answer that it is OK to move it to Undeclared.  If you're bootstrapping Croquet 1.0 expect one compilation error in CTerrainTool>>build.  It uses a future which isn't sent a message.

Once you have run the bootstrap read and obey the comment in the pop-up window as part of which you will file-in cleanup.st.  Good luck!!"

| sic previouslyUndeclared newlyUndeclared hasTweak isSqueak39 getLatest csdir loadCS canAndShouldDeleteGenerateWithUsing ctx |
#(initFields rmoveAppNamed: startCallbackProcess startLoggerProcess voidMapCache) size. "Squash compiler warnings"

(sic := SmalltalkImage current)
	closeSourceFiles; "so copying the changes file will always work"
	saveChangesInFileNamed: (sic fullNameForChangesNamed: 'ClosureBootstrap.image');
	saveImageInFileNamed: (sic fullNameForImageNamed: 'ClosureBootstrap.image').

ctx := thisContext.
[ctx == nil or:[ctx receiver isKindOf: FileStream]] whileFalse:[ctx := ctx sender].
ctx ifNil:[^self error: 'Can''t find install location'].
csdir := ctx receiver directory directoryNamed: 'changesets'.

Undeclared isEmpty ifFalse:
	[Undeclared removeUnreferencedKeys].
previouslyUndeclared := Undeclared keys.

hasTweak := Smalltalk includesKey: #Tweak.
canAndShouldDeleteGenerateWithUsing := (MethodNode includesSelector: #generateWith:using:)
										and: [(SystemNavigation default allCallsOn: #generateWith:using:) size <= 1].
isSqueak39 := Parser includesSelector: #pragmaSequence.

getLatest := [:csn| | files splits ext |
	ext := '.cs'.
	files := csdir fileNamesMatching: csn, '*', ext.
	files isEmpty ifTrue: [ext := '.st'. files := csdir fileNamesMatching: csn, '*', ext].
	splits := files
			collect: [:file | csdir splitNameVersionExtensionFor: file]
			thenSelect: [:split | (split at: 1) = csn].
	splits := splits asSortedCollection: [:a :b | (a at: 2) < (b at: 2)].
	csn, ((splits last at: 2) = 0 ifTrue: [''] ifFalse: ['.', (splits last at: 2) asString]), ext].

loadCS := [:csn| "File-in a changeset.  FIlter out all Closure Usage Fixes.  These
				are redeclarations of block-local temps from method-level to block-local scope.
				We want these fixes but they're too various and hence too risky.  Better to
				hand-roll them specific to a given release."
			(csn beginsWith: 'Closure Usage Fixes')
				ifTrue:
					[Transcript cr; nextPutAll: 'Filtered out: '; nextPutAll: csn; cr; flush.]
				ifFalse:
					[Transcript cr; nextPutAll: csn; cr; flush.
					(csdir fileNamed: (getLatest value: csn)) fileIn]].
#(	'compiler-preload'
	'NeededFixes'
	'Debugger PC ParseNode Fixes'
	'PreRecompilationPatches'
	'DebuggerMethodMap'
	'MergeCompilers') do: loadCS.
hasTweak ifTrue:
	[loadCS value: 'MergeCompilers-TweakSide'.
	(Smalltalk includesKey: #CAssignmentTile) ifTrue:
		[loadCS value: 'MergeCompilers-CroquetSide'].
	(Smalltalk includesKey: #TScriptContext) ifTrue:
		[loadCS value: 'MergeCompilers-QwaqSide'].
	loadCS value: 'Recompilation-TweakSide'].
isSqueak39 ifTrue:
	[loadCS value: 'Recompilation-Squeak39Side'].
#(	'Recompilation'
	'BytecodeBrowsing'
	'Married Context Management Access'
	'Exeunt Old BlockClosure'
	'ParseNodeVisitorPattern'
	'Closure Compiler'
	'Closure Usage Fixes'
	'Closure Compiler Tests'
	'BlockContext Usage Fixes'
	'NoContextSwitchValue') do: loadCS.
isSqueak39 ifTrue:
	[loadCS value: 'BlockContext Usage Fixes-Squeak39Side'.
	 loadCS value: 'Closure Compiler-Squeak39Side'].
hasTweak
	ifTrue:
		[loadCS value: 'Closure Compiler-TweakSide'.
		loadCS value: 'Tweak-Closure-Extensions'.
		(Smalltalk includesKey: #ScriptProcess) ifTrue:
			[loadCS value: 'Closure Usage Fixes-CroquetSide']]
	ifFalse:
		[loadCS value: 'Recompilation-SqueakSide'].
(Smalltalk includesKey: #BufferedAsyncFileReadStream) ifTrue:
	[loadCS value: 'Closure Usage Fixes-BARFS'].
canAndShouldDeleteGenerateWithUsing ifTrue:
	[(SystemNavigation default allCallsOn: #generateWith:using:) isEmpty
		ifTrue: [MethodNode removeSelector: #generateWith:using:]
		ifFalse: [nil inform: 'Did not remove MethodNode>>#generateWith:using:.  It has senders']].
Undeclared removeUnreferencedKeys.
(Undeclared keys anySatisfy: [:k| (previouslyUndeclared includes: k) not]) ifTrue:
	[nil inform: 'Undeclared has acquired some entries during the bootstrap.'.
	 newlyUndeclared := Dictionary new.
	 Undeclared associations do:
		[:assoc|
		(previouslyUndeclared includes: assoc key) ifFalse:
			[newlyUndeclared add: assoc]].
	 newlyUndeclared inspect].

"Now recompile using closures"
loadCS value: 'Compile Using Closures'.

"Now start to clean-up."
WeakArray restartFinalizationProcess.
MethodChangeRecord allInstancesDo:[:x| x noteNewMethod: nil].
Undeclared removeUnreferencedKeys.
Delay startTimerInterruptWatcher.
WorldState allInstancesDo:[:ws| ws convertAlarms; convertStepList].
(Workspace canUnderstand: #initializeBindings) ifTrue:
	[Workspace allInstancesDo:
		[:ws| ws initializeBindings]].
ExternalDropHandler initialize.
ScrollBar initializeImagesCache.
Vocabulary initialize.
Smalltalk garbageCollect.
GradientFillStyle initPixelRampCache.
Smalltalk
	at: #SokobanMorph
	ifPresent: [:sm| sm initFields].
Smalltalk
	at: #DebuggerMethodMap
	ifPresent: [:dmm| dmm voidMapCache].
Smalltalk
	at: #KClipboard
	ifPresent: [:kcb| kcb clearDefault].
Smalltalk
	at: #ServiceRegistry
	ifPresent: [:sr| sr rebuild].
(ProcessBrowser respondsTo: #registerWellKnownProcesses) ifTrue:
	[ProcessBrowser registerWellKnownProcesses].
#(Python QFileDialog) do:
	[:className |
	Smalltalk
		at: className
		ifPresent:[:class| class stopCallbackProcess]].
Smalltalk
	at: #TEmbeddedApp
	ifPresent:
		[:tea|
		(tea respondsTo: #removeObsolete)
			ifTrue: [tea removeObsolete]
			ifFalse: [(tea classPool at: #RegisteredApps) associationsDo:
						[:a| a value isObsolete ifTrue: [tea rmoveAppNamed: "sic" a key]]]].
Smalltalk
	at: #TErrorReportPlugin
	ifPresent: [:terp| | p |
	 (p := terp classPool at: #LoggerProcess ifAbsent: [nil]) ifNotNil:
		[p isTerminated
			ifTrue: [terp classPool at: #LoggerProcess put: nil]
			ifFalse: [terp startLoggerProcess]]].
Smalltalk
	at: #DebuggerMethodMap
	ifPresent: [:dmm| dmm voidMapCache].

UIManager default edit: 'The closure installation is complete. Now do the following:
	* Close any open file lists, browsers, message names etc.
	* Close all Monticello browsers.
	* Hit Alt-Period (to terminate the current Morphic process) and close the notifier.
	* Save and quit, then restart the image.
Once you have done all this, file in cleanup.st to clean up all unbound methods and
obsolete block contexts.'.
